from .ioc import SingleObjectContainer
from .module import Module
import sys


class ModuleExecuter(Module):
    """
    Simple executer which parses kwargs and args. Passes kwargs into itself, and calls "start"
    """
    kwargs = None
    args = None
    exit_code = 0

    def __init__(self):
        self.entry()

    def register(self, module):
        if self.__ioc__ is None:
            self.__ioc__ = SingleObjectContainer(obj_type=Module)
        module.__ioc__ = self.__ioc__
        self.__ioc__.register(module.id, module)

    # Entry point for the executer
    def entry(self):
        self.kwargs, self.args = parse_sysargs()
        self.start(kwargs=self.kwargs)


def parse_sysargs():
    current_key = None
    kwargs = dict()
    args = list()
    try:
        for arg in sys.argv[1:]:
            if current_key is None:
                if arg.startswith('-'):
                    current_key = arg.lstrip('-')
                else:
                    args.append(arg)
                continue
            if current_key is not None and not arg.startswith("-"):
                kwargs[current_key] = arg
                current_key = None
            else:
                kwargs[current_key] = None
                current_key = arg.lstrip('-')
        if current_key is not None and current_key not in kwargs:
            kwargs[current_key] = None
            current_key = None
    except IndexError:
        args.append(sys.argv[1])

    return kwargs, args



"""
cf_stack.py

CloudFormation Stack help methods - currently extending boto3

Note:
    Credentials are required to communicate with AWS.
    aws cli profile can be passed in using --profile, or
    the following ENVIRONMENT VARIABLES can be set before
    running this script:
        AWS_ACCESS_KEY_ID
        AWS_SECRET_ACCESS_KEY
"""

import sys
import os
import boto3
import botocore
import argparse
import json
import yaml
import logging
import time

logging.getLogger("botocore").setLevel(logging.CRITICAL)


def read_from_file(file_path):
    '''
    Read from a file and return a dict
    :param file_path: path to file
    :return: dict
    '''
    template_body = {}
    if os.path.exists(file_path):
        with open(file_path, 'r') as stream:
            if file_path.endswith('.yaml'):
                template_body = yaml.load(stream, yaml.SafeLoader)
            else:
                # Assume json if not yaml
                template_body = json.load(stream)
    else:
        logging.error('given file: ' + file_path + ' does not exist')
    return template_body


def process_stack_params_arg(stack_params):
    '''
    stack_params should be a list of key=value pairs
    :param stack_params: list of 'key=value' strings
    :return: list of dicts
    '''
    stack_parameters = []
    for param in stack_params:
        key, value = param.split('=')
        stack_parameters.append({'ParameterKey': key, 'ParameterValue': value})
    return stack_parameters


def query_stack_status_in_region(region, stack_name, profile=None):
    session = boto3.session.Session(profile_name=profile, region_name=region)
    cf_client = session.client('cloudformation')
    result = None
    try:
        query = cf_client.describe_stacks(StackName=stack_name)
        result = query['Stacks'][0]['StackStatus']
    except botocore.exceptions.ClientError as e:
        if e.response['Error']['Code'] == 'ValidationError':
            # Stack doesn't exist - set create to True
            result = 'DOES_NOT_EXIST'
        else:
            logging.error('Unexpected error: %s' % e)
    return result


def query_stack_status(region_list, stack_name, profile=None):
    result = {}

    for region in region_list:
        logging.debug("Querying stack in region: " + region)
        result[region] = query_stack_status_in_region(region, stack_name, profile=profile)
    return result


def get_stack_events_in_region(region, stack_name, profile=None):
    session = boto3.session.Session(profile_name=profile, region_name=region)
    cf_client = session.client('cloudformation')
    events = []
    try:
        current_event_set = cf_client.describe_stack_events(StackName=stack_name)
        events.extend(current_event_set['StackEvents'])
        next_token = None
        if 'NextToken' in current_event_set:
            next_token = current_event_set['NextToken']
        while next_token:
            current_event_set = cf_client.describe_stack_events(StackName=stack_name, NextToken=next_token)
            events.extend(current_event_set['StackEvents'])
            next_token = None
            if 'NextToken' in current_event_set:
                next_token = current_event_set['NextToken']
    except botocore.exceptions.ClientError as e:
        logging.error("Connection error to AWS. Check your credentials. Error: %s" % e)
    return events


def get_stack_events(region_list, stack_name, profile=None):
    result = {}

    for region in region_list:
        logging.debug("Querying stack in region: " + region)
        result[region] = query_stack_status_in_region(region, stack_name, profile=profile)
    return result


def delete_stack_in_region(region, stack_name, profile=None):
    session = boto3.session.Session(profile_name=profile, region_name=region)
    cf_client = session.client('cloudformation')
    result = None
    response = cf_client.delete_stack(StackName=stack_name)
    if 'ResponseMetadata' in response and 'HTTPStatusCode' in response['ResponseMetadata']:
        result = (response['ResponseMetadata']['HTTPStatusCode'] == 200)
    return result


def delete_stack(region_list, stack_name, profile=None):
    result = {}

    for region in region_list:
        logging.debug("Deleting stack in region: " + region)
        result[region] = delete_stack_in_region(region, stack_name, profile=profile)
    return result


def update_stack_in_region(region, stack_name, stack_params, template_body, tags=None, new_stack=False, profile=None, dryrun=False):
    '''
    Update a stack in the given region
    :param region: region to create/update the stack in
    :param stack_name: name of the stack
    :param stack_params: stack parameters (list of dicts)
    :param template_body: body of the template as a dict
    :return: ARN of the stack, if created or None
    '''
    session = boto3.session.Session(profile_name=profile, region_name=region)
    cf_client = session.client('cloudformation')

    result = False

    create = False
    if new_stack:
        create = True

    # TODO: create_change_set

    # Validate the template
    logging.debug("Template body:\n" + str(template_body))
    try:
        cf_client.validate_template(TemplateBody=json.dumps(template_body))
    except botocore.exceptions.ClientError as e:
        if e.response['Error']['Code'] == 'ValidationError':
            logging.error('Validation of given template failed')
        else:
            logging.error('Unexpected error: %s' % e)
        return result

    logging.debug("Stack params:\n" + str(stack_params))

    # Make sure stack can be created/updated
    stack_status = query_stack_status_in_region(region, stack_name, profile)
    if stack_status:
        if new_stack:
            if stack_status != 'DOES_NOT_EXIST':
                logging.error('Cannot create - stack already exists - use update to update it')
                return result
        else:
            if stack_status == 'DOES_NOT_EXIST':
                # Stack doesn't exist
                create = True
            else:
                if not stack_status.endswith('_COMPLETE'):
                    logging.error('Stack is NOT in an updatable state')
                    logging.error('Current statck status is %s' % stack_status)
                    return result
    else:
        logging.error('Unable to get stack status')
        return result

    stack_arn = None

    if create:
        if tags:
            response = cf_client.create_stack(StackName=stack_name,
                                              TemplateBody=json.dumps(template_body),
                                              Parameters=stack_params,
                                              Tags=tags,
                                              Capabilities=['CAPABILITY_NAMED_IAM'])
        else:
            response = cf_client.create_stack(StackName=stack_name,
                                              TemplateBody=json.dumps(template_body),
                                              Parameters=stack_params,
                                              Capabilities=['CAPABILITY_NAMED_IAM'])
        if 'ResponseMetadata' in response and 'HTTPStatusCode' in response['ResponseMetadata'] \
                and response['ResponseMetadata']['HTTPStatusCode'] == 200:
            if 'StackId' in response:
                stack_arn = response['StackId']
    else:
        try:
            if tags:
                response = cf_client.update_stack(StackName=stack_name,
                                                  TemplateBody=json.dumps(template_body),
                                                  Parameters=stack_params,
                                                  Tags=tags,
                                                  Capabilities=['CAPABILITY_NAMED_IAM'])
            else:
                response = cf_client.update_stack(StackName=stack_name,
                                                  TemplateBody=json.dumps(template_body),
                                                  Parameters=stack_params,
                                                  Capabilities=['CAPABILITY_NAMED_IAM'])
            if 'ResponseMetadata' in response and 'HTTPStatusCode' in response['ResponseMetadata'] \
                    and response['ResponseMetadata']['HTTPStatusCode'] == 200:
                if 'StackId' in response:
                    stack_arn = response['StackId']
        except botocore.exceptions.ClientError as e:
            if e.response['Error']['Code'] == 'ValidationError' and 'No updates are to be performed' in e.response['Error']['Message']:
                logging.warn('No changes detected')
                result = True
            else:
                logging.error('Unexpected error: %s' % e)
                return result

    if stack_arn:
        # Got a stack_arn - query the status of the stack creation
        CURRENT_CHECK = 0
        MAX_CHECKS = 60
        SLEEP_SECONDS = 30
        logging.info(('Creating' if create else 'Updating') + ' stack with name %s in region %s ' % (stack_name, region))
        logging.info("*** This may take up to %5d seconds..." % (MAX_CHECKS * SLEEP_SECONDS))
        stack_status = "Unknown"
        while CURRENT_CHECK <= MAX_CHECKS:
            stack_status = query_stack_status_in_region(region, stack_name, profile)
            if 'ROLLBACK' in stack_status:
                logging.error('*** ' + ('create' if create else 'update') + ' failed')
                logging.error('*** Waiting 5 minutes to make sure stack rolled back successfully')
                time.sleep(300)
                stack_status = query_stack_status_in_region(region, stack_name, profile)
                if 'FAILED' in stack_status:
                    logging.critical("*** Rollback has failed")
                else:
                    logging.info("*** Stack rolled back")

                logging.error("***  Stack operation failed.")
                events = get_stack_events_in_region(region, stack_name, profile)
                stack_events = ""
                for event in events:
                    stack_events += "%s: %s - %s - %s\n" % (str(event['Timestamp']), event['ResourceStatus'],
                                                            event['ResourceType'], event['LogicalResourceId']) \
                                    + ((('   %s\n') % event['ResourceStatusReason']) if 'ResourceStatusReason' in event else (''))
                logging.error("Stack events:\n%s" % stack_events)
                if create:
                    # This was a new stack - remove the unstable stack
                    logging.info("*** Removing unstable stack ...")
                    delete_stack_in_region(region, stack_name, profile)
                result = False
                break
            elif 'COMPLETE' in stack_status:
                logging.info('*** ' + ('create' if create else 'update') + ' completed successfully')
                result = True
                break
            else:
                logging.info("Current stack status: %s" % stack_status)
            CURRENT_CHECK += 1
            time.sleep(SLEEP_SECONDS)
        if CURRENT_CHECK > MAX_CHECKS and 'COMPLETE' not in stack_status:
            logging.error("*** Stack has not yet stabilized in %5d seconds - check the cloudformation console or the ECS events tab for more detail" % (MAX_CHECKS * SLEEP_SECONDS))
    return result


def update_stack(region_list, stack_name, stack_params, template_body, tags=None, profile=None, dryrun=False):
    '''
    Create/Update a stack in the given regions
    :param region_list: list of regions to create the stack in
    :param stack_name: name of the stack
    :param stack_params: stack parameters (list of dicts)
    :param template_body: body of the template as a dict
    :return: ARN of the stack(s) or None if not created
    '''
    result = {}

    for region in region_list:
        logging.debug("Creating stack in region: " + region)
        result[region] = update_stack_in_region(region, stack_name, stack_params, template_body,
                                                tags=tags, profile=profile, dryrun=dryrun)
    return result


def create_stack(region_list, stack_name, stack_params, template_body, tags=None, profile=None, dryrun=False):
    '''
    Create a stack in the given regions
    :param region_list: list of regions to create the stack in
    :param stack_name: name of the stack
    :param stack_params: stack parameters (list of dicts)
    :param template_body: body of the template as a dict
    :return: ARN of the stack(s) or None if not created
    '''
    result = {}

    for region in region_list:
        logging.debug("Creating stack in region: " + region)
        result[region] = update_stack_in_region(region, stack_name, stack_params, template_body, tags=None,
                                                new_stack=True, profile=profile, dryrun=dryrun)
    return result


def _update_stack_parameters(region, stack_id, parameters, profile=None, dryrun=False):
    """
    Update a given stack with the given parameters
    :param region: region that the stack exists in
    :param stack_id: the name or ID of the stack
    :param parameters: list of parameter objects
    :param dryrun: if true, no changes are made
    :return: json object
    """
    if not region:
        logging.error("ERROR: You must supply a region to scan")
        return None
    else:
        logging.info('Updating Stack: ' + stack_id)
        for param in parameters:
            if 'PreviousValue' in param:
                logging.info('   ' + param['ParameterKey'])
                logging.info('      OLD: ' + param['PreviousValue'])
                logging.info('      NEW: ' + param['ParameterValue'])
                del param['PreviousValue']
        if not dryrun:
            stack = get_stack_with_name_or_id(region, stack_id)
            session = boto3.session.Session(profile_name=profile, region_name=region)
            cf_client = session.client('cloudformation')
            try:
                if 'Capabilities' in stack:
                    status = cf_client.update_stack(StackName=stack_id, Parameters=parameters, UsePreviousTemplate=True, Capabilities=stack['Capabilities'])['StackId']
                else:
                    status = cf_client.update_stack(StackName=stack_id, Parameters=parameters, UsePreviousTemplate=True)['StackId']
                return status
            except botocore.exceptions.ClientError as e:
                if e.response['Error']['Code'] == 'ValidationError' and 'No updates are to be performed' in \
                        e.response['Error']['Message']:
                    logging.error("   ERROR: New value matches Old value - no update required")
                else:
                    logging.error('Unexpected error: %s' % e)
    return False


def get_stacks_with_given_parameter(region, parameter_list, profile=None):
    """
    Get a list of stacks that have at least one parameter with a name in the given list
    :param region: the region to scan
    :param parameter_list: list of possible parameter names to look for
    :return: list of stacks that have a parameter with a name in the given list
    """
    stacks_with_given_parameter = []
    if not region:
        logging.error("ERROR: You must supply a region to scan")
        return None
    else:
        session = boto3.session.Session(profile_name=profile, region_name=region)
        cf_client = session.client('cloudformation')
        cf_data = cf_client.describe_stacks()
        if "Stacks" in cf_data:
            for stack in cf_data["Stacks"]:
                if "Parameters" in stack:
                    for parameter in stack["Parameters"]:
                        if parameter['ParameterKey'] in parameter_list:
                            logging.debug(
                                "Found parameter - " + parameter['ParameterKey'] + ' - in stack: ' + stack['StackName'])
                            stacks_with_given_parameter.append(stack)
                            break
    return stacks_with_given_parameter


def get_stack_with_name_or_id(region, stack_id, profile=None):
    """
    Get the stack with the given name or ID
    :param region: region where the stack exists
    :param stack_id: stack name or ID
    :return: stack in question or empty dictionary
    """
    stack = {}
    if not region:
        logging.error("ERROR: You must supply a region to scan")
        return None
    else:
        session = boto3.session.Session(profile_name=profile, region_name=region)
        cf_client = session.client('cloudformation')
        cf_data = cf_client.describe_stacks(StackName=stack_id)
        logging.debug("Getting stack description for stack with name/id: " + stack_id)
        if "Stacks" in cf_data:
            if len(cf_data["Stacks"]) > 1:
                # Problem - mutliple stacks with this name??
                logging.error("Error: Multiple stacks with given name")
            else:
                stack = cf_data['Stacks'][0]
    return stack


def get_new_parameter_list_for_update(stack, expected_value, new_value, parameter_to_change, force=False):
    """
    Given a list of possible parameter names, generate a new parameter list with
    existing_value changed to new_value
    :param stack: the stack to mine the parameter list from
    :param expected_value: expected existing value for the parameter
    :param new_value: the new value to use for the parameter
    :param parameter_to_change: list containing possible parameter names
    :return: tuple containing a boolean for whether an update is required and the new parameter list
    """
    new_parameters_list = []
    update_required = False
    for parameter in stack['Parameters']:
        new_param = {}
        new_param['ParameterKey'] = parameter['ParameterKey']
        if parameter['ParameterKey'] in parameter_to_change:
            if force or (expected_value in parameter['ParameterValue']):
                update_required = True
                new_param['ParameterValue'] = new_value
                new_param['PreviousValue'] = parameter['ParameterValue']
            else:
                logging.warn(
                    "Unexpected value detected - Stack will NOT be updated\n   Stack: " + stack[
                        'StackName'] + "\n   Existing value: " + parameter['ParameterValue'])
                new_param['UsePreviousValue'] = True
        else:
            new_param['UsePreviousValue'] = True
        new_parameters_list.append(new_param)
    return update_required, new_parameters_list


def list_stacks_with_given_parameter_in_region(region, parameter_list, profile=None):
    """
    Print a list of stacks in a given region that contain at least one parameter in the given parameter list
    :param region: region to scan
    :param parameter_list: list of possible parameter names
    """
    if not region:
        logging.error("You must supply a region to scan")
    else:
        logging.info(
            "\nCloudFormation Stacks in region: " + region + " with at least one of the following parameters: " + ', '.join(
                parameter_list))
        stacks_with_given_parameter = get_stacks_with_given_parameter(region, parameter_list, profile=profile)
        if len(stacks_with_given_parameter) > 0:
            for stack in stacks_with_given_parameter:
                for parameter in stack["Parameters"]:
                    if parameter['ParameterKey'] in parameter_list:
                        logging.info(
                            "Stack Name: " + stack["StackName"] + "\n   " + parameter['ParameterKey'] + ": " +
                            parameter[
                                "ParameterValue"])
                        break
        else:
            logging.info("None")


def list_stacks_with_given_parameter(region_list, parameter_list, profile=None):
    """
    Print a list of stacks for the given regions that contain at least one parameter in the given parameter list
    :param region_list: list of regions
    :param parameter_list: list of possible parameter names
    """
    for region in region_list:
        list_stacks_with_given_parameter_in_region(region, parameter_list, profile=profile)


def update_stack_with_given_parameter(region, stack_name, expected_value, new_value, parameter_to_change, profile=None,
                                      dryrun=False, force=False):
    """
    Update the given parameter in the given stack (for the given region) to a new value, provided
    the existing value contains the expected_value
    :param region: region where the stack exists
    :param stack_name: the stack name or ID
    :param expected_value: the existing value of the parameter should contain the expected_value
    :param new_value: the new value for the parameter
    :param parameter_to_change: list of possible names for the parameter
    :param dryrun: if true, no changes are made
    :returns: True / False
    """
    stack = get_stack_with_name_or_id(region, stack_name)
    update_required, new_parameter_list = get_new_parameter_list_for_update(stack, expected_value, new_value,
                                                                            parameter_to_change, force)
    if update_required:
        return _update_stack_parameters(region, stack["StackId"], new_parameter_list, profile=profile, dryrun=dryrun)


def update_all_stacks_with_given_parameter(region, expected_value, new_value, parameter_to_change, profile=None,
                                           dryrun=False, force=False):
    """
    Update the given parameter in all stacks (for the given region) to a new value, provided
    the existing value contains the expected_value
    :param region: region where the stacks exist
    :param expected_value: the existing value of the parameter should contain the expected_value
    :param new_value: the new value for the parameter
    :param parameter_to_change: list of possible names for the parameter
    :param dryrun: if true, no changes are made
    """
    logging.info('\nUpdating all matching Stacks in region: ' + region)
    update_status = {}
    for stack in get_stacks_with_given_parameter(region, parameter_to_change, profile=profile):
        update_required, new_parameter_list = get_new_parameter_list_for_update(stack, expected_value, new_value,
                                                                                parameter_to_change, force)
        if update_required:
            update_status[stack['StackId']] = _update_stack_parameters(region, stack["StackId"], new_parameter_list,
                                                                       profile=profile, dryrun=dryrun)

    return update_status


if __name__ == "__main__":

    parser = argparse.ArgumentParser(
        description='Script to view/modify CloudFormation Stacks')

    me_cmd_group = parser.add_mutually_exclusive_group(required=True)
    me_cmd_group.add_argument("--query-status", help="query stack status in given region(s)", dest='query_stack', metavar='STACK_NAME')
    me_cmd_group.add_argument("--delete-stack", help="delete stack in given region(s)", dest='delete_stack', metavar='STACK_NAME')
    me_cmd_group.add_argument("--create-stack", help="create stack in given region(s)", dest='create_stack', metavar='STACK_NAME')
    me_cmd_group.add_argument("--update-stack", help="update stack in given region(s)", dest='update_stack', metavar='STACK_NAME')
    me_cmd_group.add_argument("--update",
                              help="Update Parameter to new value for given stack in the specified region - must supply expected existing value and new value. STACK_ID can be the name or ID of the Stack",
                              dest='update', nargs=3, metavar=('STACK_ID', 'EXPECTED_VALUE', 'NEW_VALUE'))
    me_cmd_group.add_argument("--update-all",
                              help="Update Parameter to new value for all stacks in the specified region - must supply expected existing value and new value",
                              dest='update_all', nargs=2, metavar=('EXPECTED_VALUE', 'NEW_VALUE'))
    me_cmd_group.add_argument("--list",
                              help="List all stacks in given region(s) that have a given parameter",
                              dest='list', action='store_true')

    parser.add_argument("--stack-params", help="space separated list of key=value stack parameters", dest='stack_params',
                        nargs='+', required=False)
    parser.add_argument("--template-body", help="CFN template body (as json/yaml - preface with file:// if file)", dest='template_body',
                        required=False)
    parser.add_argument("--param", help="space separated list of possible names for the parameter", dest='param',
                        nargs='+', required=False)
    parser.add_argument("--regions", help="Specify regions (space separated)", dest='regions', nargs='+', required=True)
    parser.add_argument("--profile",
                        help="The name of an aws cli profile to use.", dest='profile', required=False)
    parser.add_argument("--dryrun", help="Do a dryrun - no changes will be performed", dest='dryrun',
                        action='store_true', default=False, required=False)
    parser.add_argument("--force", help="Force an update, even if expected value doesn't match", dest='force',
                        action='store_true', default=False, required=False)
    parser.add_argument("--verbose", help="Turn on DEBUG logging", action='store_true', required=False)
    args = parser.parse_args()

    log_level = logging.INFO

    if args.verbose:
        log_level = logging.DEBUG

    logger = logging.getLogger()
    logger.setLevel(logging.DEBUG)
    # create file handler which logs even debug messages
    fh = logging.FileHandler('cf_stack.log')
    fh.setLevel(logging.DEBUG)
    # create console handler using level set in log_level
    ch = logging.StreamHandler()
    ch.setLevel(log_level)
    console_formatter = logging.Formatter('%(levelname)8s: %(message)s')
    ch.setFormatter(console_formatter)
    file_formatter = logging.Formatter('%(asctime)s - %(levelname)8s: %(message)s')
    fh.setFormatter(file_formatter)
    # Add the handlers to the logger
    logger.addHandler(fh)
    logger.addHandler(ch)

    logging.debug("INIT")

    if args.dryrun:
        logging.info("***** Dryrun selected - no changes will be made *****")

    if args.regions:
        logging.info("Regions: " + str(args.regions))

    if not args.param:
        if args.list or args.update or args.update_all:
            logging.critical("Must supply a parameter")
            sys.exit(1)

    if args.create_stack or args.update_stack:
        if not args.template_body:
            logging.critical("Must supply a template body")
            sys.exit(1)

        if args.template_body.startswith('file://'):
            # Template is a file - read it in
            file_path = args.template_body[7:]
            template_body = read_from_file(file_path)
            if not template_body:
                sys.exit(1)
        else:
            template_body = args.template_body

        stack_params = process_stack_params_arg(args.stack_params)

        result = None
        if args.create_stack:
            result = create_stack(args.regions, args.create_stack, stack_params, template_body, args.profile, args.dryrun)
        elif args.update_stack:
            result = update_stack(args.regions, args.update_stack, stack_params, template_body, args.profile, args.dryrun)

        if not args.dryrun:
            if result:
                logging.info("Create/Update results:")
                for region in result:
                    logging.info('   ' + region + ': ' + ("Success" if result[region] else "Failed"))
            else:
                logging.error("Create/Update failed.")

    elif args.delete_stack:
        result = delete_stack(args.regions, args.delete_stack, args.profile)

    elif args.query_stack:
        result = query_stack_status(args.regions, args.query_stack, args.profile)

    elif args.list:
        if not args.param:
            logging.critical("Must supply a parameter name")
            sys.exit(1)
        logging.info("Searching for Stacks with a Parameter with name(s): " + ' '.join(args.param))
        list_stacks_with_given_parameter(args.regions, args.param, args.profile)

    elif args.update:
        if 'all' in args.regions:
            logging.critical("Only one region can be specified with update", "error")
            sys.exit(1)
        if not args.param:
            logging.critical("Must supply a parameter name")
            sys.exit(1)
        update_status = update_stack_with_given_parameter(args.regions, args.update[0], args.update[1], args.update[2],
                                                          args.param, args.profile, args.dryrun, args.force)
        if not args.dryrun:
            if update_status:
                logging.info("Stack Parameter Update Succeeded")
            else:
                logging.error("Stack Parameter Update Failed")

    elif args.update_all:
        if len(args.regions) > 1:
            logging.critical("Only one region can be specified with update", "error")
            sys.exit(1)
        if not args.param:
            logging.critical("Must supply a parameter name")
            sys.exit(1)
        update_status = update_all_stacks_with_given_parameter(args.regions, args.update_all[0], args.update_all[1],
                                                               args.param, args.profile, args.dryrun, args.force)
        if not args.dryrun and update_status:
            logging.info("UPDATE STATUS:")
            for stack in update_status:
                logging.info(stack + ": " + ('Succeeded' if update_status[stack] else 'Failed'))



"""
dynamodb.py

AWS Dynamo DB helper methods - currently extending boto3

Note:
    Credentials are required to communicate with AWS.
    aws cli profile can be passed in using --profile, or
    the following ENVIRONMENT VARIABLES can be set before
    running this script:
        AWS_ACCESS_KEY_ID
        AWS_SECRET_ACCESS_KEY
"""

import logging
import argparse
import sys
import boto3
import botocore


def get_all_tables_in_region(region, profile=None):
    if not region:
        logging.error("You must supply a region")
        return []

    session = boto3.session.Session(profile_name=profile, region_name=region)
    dynamodb = session.client('dynamodb')

    all_tables = []
    try:
        current_set = dynamodb.list_tables()
        all_tables.extend(current_set['TableNames'])
        while 'LastEvaluatedTableName' in current_set:
            start_name = current_set['LastEvaluatedTableName']
            current_set = dynamodb.list_tables(ExclusiveStartTableName=start_name)
            all_tables.extend(current_set['TableNames'])
    except botocore.exceptions.ClientError as e:
        logging.error('Unexpected error: %s' % e)
        raise e

    return all_tables


def get_all_tables(region_list, profile=None):
    result = {}

    for region in region_list:
        logging.debug("Checking region: " + region)
        result[region] = get_all_tables_in_region(region, profile)
    return result


def table_exists_in_region(region, table_name, profile=None, suppress_warning=False):
    if not region:
        logging.error("You must supply a region")
        return []

    session = boto3.session.Session(profile_name=profile, region_name=region)
    dynamodb = session.client('dynamodb')

    result = False
    try:
        response = dynamodb.describe_table(TableName=table_name)
        if 'Table' in response:
            result = True
    except botocore.exceptions.ClientError as e:
        if e.response['Error']['Code'] == 'ResourceNotFoundException':
            if not suppress_warning:
                logging.warn("The given table %s does not exist in region %s" % (table_name, region))
        else:
            logging.error('Unexpected error: %s' % e)
            raise e

    return result


def table_exists(region_list, table_name, profile=None, suppress_warning=False):
    result = {}

    for region in region_list:
        logging.debug("Checking region: " + region)
        result[region] = table_exists_in_region(region, table_name, profile, suppress_warning)
    return result


def get_all_items_in_table_in_region(region, table_name, profile=None):
    if not region:
        logging.error("You must supply a region")
        return []

    all_items = []
    # TODO: Implement this
    logging.error("Not yet implemented")

    return all_items


def get_all_items_in_table(region_list, table_name, profile=None):
    result = {}

    for region in region_list:
        logging.debug("Checking region: " + region)
        result[region] = get_all_items_in_table_in_region(region, table_name, profile)
    return result


def get_item_from_table_in_region(region, table_name, key, profile=None):
    if not region:
        logging.error("You must supply a region")
        return False

    session = boto3.session.Session(profile_name=profile, region_name=region)
    dynamodb = session.client('dynamodb')

    result = None

    try:
        response = dynamodb.get_item(TableName=table_name, Key=key)
        if 'ResponseMetadata' in response:
            if 'HTTPStatusCode' in response['ResponseMetadata']:
                if response['ResponseMetadata']['HTTPStatusCode'] == 200:
                    if 'Item' in response:
                        result = response['Item']
    except botocore.exceptions.ClientError as e:
        if e.response['Error']['Code'] == 'ValidationException':
            logging.error("The provided key element(s) do not match the schema")
        else:
            logging.error('Unexpected error: %s' % e)
        raise e

    return result


def get_item_from_table(region_list, table_name,
                        item_partition_key_value, item_partition_key_name, item_partition_key_type='S',
                        item_sort_key_value=None, item_sort_key_name=None, item_sort_key_type='S',
                        profile=None):
    result = {}

    # Build the key dict
    key = {}
    key[item_partition_key_name] = {item_partition_key_type: item_partition_key_value}
    if item_sort_key_value:
        key[item_sort_key_name] = {item_sort_key_type: item_sort_key_value}

    for region in region_list:
        logging.debug("Checking region: " + region)
        if table_exists_in_region(region, table_name, profile):
            result[region] = get_item_from_table_in_region(region, table_name, key, profile)
        else:
            result[region] = None
    return result


def put_item_in_table_in_region(region, table_name, key, dryrun=False, profile=None):
    if not region:
        logging.error("You must supply a region")
        return False

    result = False
    # TODO: Implement this
    logging.error("Not yet implemented")

    return result


def put_item_in_table(region_list, table_name,
                      item_partition_key_value, item_partition_key_name, item_partition_key_type='S',
                      item_sort_key_value=None, item_sort_key_name=None, item_sort_key_type='S',
                      dryrun=False, profile=None):
    result = {}

    # Build the key dict
    key = {}
    key[item_partition_key_name] = {item_partition_key_type: item_partition_key_value}
    if item_sort_key_value:
        key[item_sort_key_name] = {item_sort_key_type: item_sort_key_value}

    for region in region_list:
        logging.debug("Checking region: " + region)
        if table_exists_in_region(region, table_name, profile):
            result[region] = put_item_in_table_in_region(region, table_name, key, dryrun, profile)
        else:
            result[region] = False
    return result


def update_item_in_table_in_region(region, table_name, key, dryrun=False, profile=None):
    if not region:
        logging.error("You must supply a region")
        return False

    result = False
    # TODO: Implement this
    logging.error("Not yet implemented")

    return result


def update_item_in_table(region_list, table_name,
                         item_partition_key_value, item_partition_key_name, item_partition_key_type='S',
                         item_sort_key_value=None, item_sort_key_name=None, item_sort_key_type='S',
                         dryrun=False, profile=None):
    result = {}

    # Build the key dict
    key = {}
    key[item_partition_key_name] = {item_partition_key_type: item_partition_key_value}
    if item_sort_key_value:
        key[item_sort_key_name] = {item_sort_key_type: item_sort_key_value}

    for region in region_list:
        logging.debug("Checking region: " + region)
        if table_exists_in_region(region, table_name, profile):
            result[region] = update_item_in_table_in_region(region, table_name, key, dryrun, profile)
        else:
            result[region] = False
    return result


def create_item_in_table_in_region(region, table_name, key, dryrun=False, profile=None):
    if not region:
        logging.error("You must supply a region")
        return False

    result = False
    # TODO: Implement this
    logging.error("Not yet implemented")

    return result


def create_item_in_table(region_list, table_name,
                         item_partition_key_value, item_partition_key_name, item_partition_key_type='S',
                         item_sort_key_value=None, item_sort_key_name=None, item_sort_key_type='S',
                         dryrun=False, profile=None):
    result = {}

    # Build the key dict
    key = {}
    key[item_partition_key_name] = {item_partition_key_type: item_partition_key_value}
    if item_sort_key_value:
        key[item_sort_key_name] = {item_sort_key_type: item_sort_key_value}

    for region in region_list:
        logging.debug("Checking region: " + region)
        if table_exists_in_region(region, table_name, profile):
            result[region] = create_item_in_table_in_region(region, table_name, key, dryrun, profile)
        else:
            result[region] = False
    return result


def delete_item_in_table_in_region(region, table_name, key, dryrun=False, profile=None):
    if not region:
        logging.error("You must supply a region")
        return False

    result = False
    # TODO: Implement this
    logging.error("Not yet implemented")

    return result


def delete_item_in_table(region_list, table_name,
                         item_partition_key_value, item_partition_key_name, item_partition_key_type='S',
                         item_sort_key_value=None, item_sort_key_name=None, item_sort_key_type='S',
                         dryrun=False, profile=None):
    result = {}

    # Build the key dict
    key = {}
    key[item_partition_key_name] = {item_partition_key_type: item_partition_key_value}
    if item_sort_key_value:
        key[item_sort_key_name] = {item_sort_key_type: item_sort_key_value}

    for region in region_list:
        logging.debug("Checking region: " + region)
        if table_exists_in_region(region, table_name, profile):
            result[region] = create_item_in_table_in_region(region, table_name, key, dryrun, profile)
        else:
            result[region] = False
    return result


def item_exists_in_region(region, table_name, key, profile=None):
    if get_item_from_table_in_region(region, table_name, key, profile):
        return True
    else:
        return False


def item_exists(region_list, table_name, item_partition_key_value, item_partition_key_name, item_partition_key_type='S',
                item_sort_key_value=None, item_sort_key_name=None, item_sort_key_type='S', profile=None):
    result = {}

    # Build the key dict
    key = {}
    key[item_partition_key_name] = {item_partition_key_type: item_partition_key_value}
    if item_sort_key_value:
        key[item_sort_key_name] = {item_sort_key_type: item_sort_key_value}

    for region in region_list:
        logging.debug("Checking region: " + region)
        if table_exists_in_region(region, table_name, profile):
            result[region] = item_exists_in_region(region, table_name, key, profile)
        else:
            result[region] = False
    return result


def create_table_in_region(region, table_name, dryrun=False, profile=None):
    if not region:
        logging.error("You must supply a region")
        return False

    result = False
    # TODO: Implement this
    logging.error("Not yet implemented")

    return result


def create_table(region_list, table_name, dryrun=False, profile=None):
    result = {}
    for region in region_list:
        logging.debug("Checking region: " + region)
        if not table_exists_in_region(region, table_name, profile, suppress_warning=True):
            result[region] = create_table_in_region(region, table_name, dryrun, profile)
        else:
            logging.warn("Table %s already exists in region %s" % (table_name, region))
    return result


def delete_table_in_region(region, table_name, dryrun=False, profile=None):
    if not region:
        logging.error("You must supply a region")
        return False

    result = False
    # TODO: Implement this
    logging.error("Not yet implemented")

    return result


def delete_table(region_list, table_name, dryrun=False, profile=None):
    result = {}
    for region in region_list:
        logging.debug("Checking region: " + region)
        if table_exists_in_region(region, table_name, profile):
            result[region] = delete_table_in_region(region, table_name, dryrun, profile)
        else:
            result[region] = False
    return result


if __name__ == "__main__":
    parser = argparse.ArgumentParser(description='dynamodb.py')

    me_cmd_group = parser.add_mutually_exclusive_group(required=True)
    me_cmd_group.add_argument("--get", help="Perform a Get on an item", action="store_true")
    # me_cmd_group.add_argument("--put", help="Perform a Put on an item", action="store_true")
    # me_cmd_group.add_argument("--update", help="Perform an Update on an item", action="store_true")
    # me_cmd_group.add_argument("--list-all", help="List all items in the given table and regions", action="store_true", dest="list_all_items")
    # me_cmd_group.add_argument("--delete", help="Perform a Delete on an item/table", action="store_true")
    me_cmd_group.add_argument("--exists", help="Check if an item/table exists", action="store_true")
    # me_cmd_group.add_argument("--create", help="Create a table", action="store_true")
    me_cmd_group.add_argument("--list-all-tables", help="List all tables in the given regions", action="store_true", dest="list_all_tables")

    parser.add_argument("--table", help="Table name", dest='table', required=False)
    parser.add_argument("--item", help="Item name", dest='item', required=False)
    parser.add_argument("--pkey", help="Partition key", dest='pkey', required=False)
    parser.add_argument("--pkey-type", help="Partition key type", dest='pkey_type', default='S', required=False)
    parser.add_argument("--skey", help="Sort key", dest='skey', required=False)
    parser.add_argument("--skey-type", help="Sort key type", dest='skey_type', default='S', required=False)
    parser.add_argument("--skey-value", help="Sort key value", dest='skey_value', required=False)

    parser.add_argument("--regions", help="AWS Region(s) involved (space separated)", dest='regions', nargs='+', required=True)
    parser.add_argument("--profile",
                        help="The name of an aws cli profile to use.", dest='profile', required=False)
    parser.add_argument("--verbose", help="Turn on DEBUG logging", action='store_true', required=False)
    parser.add_argument("--dryrun", help="Do a dryrun - no changes will be performed", dest='dryrun',
                        action='store_true', default=False,
                        required=False)

    args = parser.parse_args()

    log_level = logging.INFO

    if args.verbose:
        print("Verbose logging selected")
        log_level = logging.DEBUG

    # Turn down logging for botocore
    logging.getLogger("botocore").setLevel(logging.CRITICAL)

    logger = logging.getLogger()
    logger.setLevel(logging.DEBUG)
    # create file handler which logs even debug messages
    fh = logging.FileHandler('dynamodb.log')
    fh.setLevel(logging.DEBUG)
    # create console handler using level set in log_level
    ch = logging.StreamHandler()
    ch.setLevel(log_level)
    console_formatter = logging.Formatter('%(levelname)8s: %(message)s')
    ch.setFormatter(console_formatter)
    file_formatter = logging.Formatter('%(asctime)s - %(levelname)8s: %(message)s')
    fh.setFormatter(file_formatter)
    # Add the handlers to the logger
    logger.addHandler(fh)
    logger.addHandler(ch)

    logging.debug("INIT")

    if not args.list_all_tables and not args.table:
        logger.error("Must supply a table")
        sys.exit(1)

    if args.item:
        if not args.pkey:
            logger.error("Must supply a partition key")
            sys.exit(1)
        if args.skey and not args.skey_value:
            logger.error("Sort key name present, but no sort key value provided")
            sys.exit(1)
        if args.skey_value and not args.skey:
            logging.warn("Sort key value present, but no sort key provided - will be ignored")

    # Item operations
    if args.get:
        if not args.item:
            logger.error("Must supply an item to get")
            sys.exit(1)

        result = get_item_from_table(args.regions, args.table,
                                     args.item, args.pkey, args.pkey_type,
                                     args.skey_value, args.skey, args.skey_type,
                                     args.profile)
        for region in result:
            print(region + ': ' + ('\n' + str(result[region]) if result[region] else "Not Present"))

    # if args.put:
    #     # TODO: Implement this above in put_item_in_region
    #     result = put_item_in_table(args.regions, args.table,
    #                                args.item, args.pkey, args.pkey_type,
    #                                args.skey_value, args.skey, args.skey_type,
    #                                args.dryrun, args.profile)
    #     for region in result:
    #         print(region + ': ' + ('\n' + 'Success' if result[region] else "Failed"))
    #
    # if args.update:
    #     # TODO: Implement this above in update_item_in_region
    #     result = update_item_in_table(args.regions, args.table,
    #                                   args.item, args.pkey, args.pkey_type,
    #                                   args.skey_value, args.skey, args.skey_type,
    #                                   args.dryrun, args.profile)
    #     for region in result:
    #         print(region + ': ' + ('\n' + 'Success' if result[region] else "Failed"))
    #
    # # Item/Table operations
    # if args.delete:
    #     if args.item:
    #         # Item delete
    #         # TODO: Implement this above in create_item_in_table_in_region
    #         result = delete_item_in_table(args.regions, args.table,
    #                                       args.item, args.pkey, args.pkey_type,
    #                                       args.skey_value, args.skey, args.skey_type,
    #                                       args.dryrun, args.profile)
    #     else:
    #         # Table delete
    #         # TODO: Implement this above in create_table_in_region
    #         result = delete_table(args.regions, args.table, args.dryrun, args.profile)
    #     for region in result:
    #         print(region + ': ' + ('\n' + 'Success' if result[region] else "Failed"))

    if args.exists:
        if args.item:
            # Item exists
            result = item_exists(args.regions, args.table,
                                 args.item, args.pkey, args.pkey_type,
                                 args.skey_value, args.skey, args.skey_type,
                                 args.profile)
        else:
            # Table exists
            if args.pkey or args.skey or args.skey_value:
                logging.warn("Ignoring extraneous information provided")
            result = table_exists(args.regions, args.table, args.profile, suppress_warning=True)
        for region in result:
            print(region + ': ' + str(result[region]))

    # Table operations
    # if args.list_all_items:
    #     # TODO: Implement this above in create_table_in_region
    #     result = get_all_items_in_table(args.regions, args.table, args.profile)
    #     for region in result:
    #         logger.info(region + ' - ' + str(len(result[region])) + ' table' + ('' if len(result[region]) == 1 else 's') + ' present')
    #         if len(result[region]) > 0:
    #             for table in result[region]:
    #                 logger.info('   ' + table)
    #
    # if args.create:
    #     if args.item:
    #         logger.warn("--item present with --create operation; Please use --put to create an item in a table")
    #     else:
    #         # Table create
    #         # TODO: Implement this above in create_table_in_region
    #         result = create_table(args.regions, args.table, args.dryrun, args.profile)
    #         for region in result:
    #             print(region + ': ' + ('\n' + 'Success' if result[region] else "Failed"))

    if args.list_all_tables:
        result = get_all_tables(args.regions, args.profile)
        for region in result:
            logger.info(region + ' - ' + str(len(result[region])) + ' table' + ('' if len(result[region]) == 1 else 's') + ' present')
            if len(result[region]) > 0:
                for table in result[region]:
                    logger.info('   ' + table)



"""
Contains some S3 helper methods and classes, currently extending boto
"""


import boto3
import os
import sys
import botocore
from botocore.handlers import disable_signing
from botocore import UNSIGNED
from botocore.client import Config
from ..core import module


def find_files(bucket, prefix, case_sensitive=True, connection=None, anonymous=True, sha256=False):
    """
    find_files will connect and return files found in bucket with prefix, all other keys are ignored.

    Optional: case_sensitive, connection

    Returns a boto3 objectCollection containing matching files, or an empty collection. Will not return any non-file keys.
    Will raise a DownloadError with an appropriate error message when unable to return a collection.
    """
    s3client = None
    if connection is None:
        try:
            connection = get_s3_connection(anonymous=False)
            s3client = boto3.client('s3')
        except Exception:
            connection = get_s3_connection(anonymous=True)
            s3client = boto3.client('s3', config=Config(signature_version=UNSIGNED))

    if anonymous:
        s3client = boto3.client('s3', config=Config(signature_version=UNSIGNED))
    else:
        s3client = boto3.client('s3')

    # Verify we can connect to remote bucket
    verify_bucket(bucket, connection=connection)

    # Connect to the remote bucket
    remote_bucket = connection.Bucket(bucket)

    # List of returned files
    files = list()

    # Look for matching files if case insensitive mode
    if not case_sensitive:
        # Iterate over objects, and append only ones that match lower case and don't end with '/'
        for obj in remote_bucket.objects.all():
            if obj.key.lower().startswith(prefix.lower()) and not obj.key.endswith("/"):
                if sha256:
                    try:
                        objsum = s3client.head_object(Bucket=bucket, Key=obj.key, ChecksumMode='ENABLED')['ResponseMetadata']['HTTPHeaders']["x-amz-checksum-sha256"]
                    except Exception:
                        raise ChecksumFailure(f"Unable head object for checksum. Please verify sha256 exists on object {bucket}/{obj.key}")
                        objsum = "unknown"
                else:
                    objsum = s3client.get_object(Bucket=bucket, Key=obj.key)["ETag"][1:-1]
                files.append((obj, objsum))
    else:  # If we're case sensitive, just use the filter
        files = remote_bucket.objects.filter(Prefix=prefix)
        sum_files = list()
        for f in files:
            if sha256:
                try:
                    objsum = s3client.head_object(Bucket=bucket, Key=f.key, ChecksumMode='ENABLED')['ResponseMetadata']['HTTPHeaders']["x-amz-checksum-sha256"]
                except Exception:
                    raise ChecksumFailure(
                        f"Unable head object for checksum. Please verify sha256 exists on object {bucket}/{obj.key}")
                    objsum = "unknown"
            else:
                objsum = s3client.get_object_attributes(Bucket=bucket, Key=f.key, ObjectAttributes=['ETag'])[1:-1]
            sum_files.append((f, objsum))
        files = sum_files

    return files


def get_s3_connection(anonymous=True):
    """
    Returns an s3 connection object. Configures anonymous access by default.
    """

    # Connect to S3
    connection = boto3.resource('s3')

    if anonymous is True:
        # Configure anonymous access
        connection.meta.client.meta.events.register('choose-signer.s3.*', disable_signing)

    return connection


def parse_s3_url(url):
    """
    Parses a url with format s3://bucket/prefix into a bucket and prefix
    """
    if not url.startswith("s3://"):
            raise ValueError("The provided URL does not follow s3://{bucket_name}/{path}")

    # Parse into bucket and prefix
    bucket = ""
    prefix = ""
    for index, char in enumerate(url[5:]):
        # Everything before this char should be the bucket name
        if char == "/":
            # Take rest of path less '/' and s3://
            prefix = url[(index + 6):]
            break
        # Build bucket string
        bucket += char

    if not bucket:
        raise ValueError("The provided URL " + str(url) + " is not valid. Please enter a URL following s3://{bucket_name}/path")

    if not prefix:
        prefix = "/"

    return bucket, prefix


def join_s3_url(prefix, *elements):
    url_builder = prefix
    for element in elements:
        if url_builder[-1] == "/":
            url_builder += element
        else:
            url_builder += "/" + element
    return url_builder


def verify_bucket(bucket_name, connection=None):
    """
    Verifies that you have read access to bucket with bucket_name and current credentials.
    """

    if connection is None:
        try:
            connection = get_s3_connection(anonymous=False)
        except Exception:
            connection = get_s3_connection(anonymous=True)

    # Verify we can connect to the bucket
    try:
        connection.meta.client.head_bucket(Bucket=bucket_name)
    except botocore.exceptions.ClientError as e:
        error_code = int(e.response['Error']['Code'])
        if error_code == 404:
            raise DownloadError("Unable to connect to remote bucket. Please verify bucket with name " + str(bucket_name) + " exists.")
        elif error_code == 403:
            raise DownloadError("Unable to connect to remote bucket. Access is forbidden.")
        else:
            raise e


class DownloadError(Exception):
    pass

class ChecksumFailure(Exception):
    pass


# MODULES #

BUCKET_KEYS = ["b", "bucket"]
CASE_INSENSITIVE_KEYS = ["i", "case-insensitive"]
DESTINATION_KEYS = ["d", "destination"]
PATH_KEYS = ["p", "prefix"]
REGION_KEYS = ["r", "region"]
SOURCE_KEYS = ["s", "source"]
HELP_KEYS = ["h", "help"]


class AsyncS3Downloader(module.AsyncModule):
        """
        AsyncS3Downloader will download a set of files identified by 'prefix' to 'destination_path' from 'bucket_name' in 'region'. You may optionally use a source URL in the form of 's3://bucket/prefix' or set 'case_insensitive' to match a path/prefix ignoring case.

        """
        HELPTEXT = """
                ----- S3 Downloader -----

The S3 Downloader will retreive a file from an S3 bucket with anonymous
read access.

-b, --bucket <bucket_name>:         Specify the name of the bucket to access
                                            (required)
-d, --destination <dest_path>       Specify the destination on local filesystem
                                            (default  './')
-i, --case-insensitive:             Specify if the path should be treated as case insensitive
                                            (default case sensitive)
-p, --prefix <path>:                Specify the path prefix within the bucket to access
                                            (default '/")
-r, --region <aws_region>:          Specify the amazon region to locate the bucket
                                            (default 'us-east-1')
-s, --source <src_url>:             Specify the source URL
                                            (ignored when bucket is set)
-h, --help:                         Display this help text

"""

        bucket_name = None
        case_insensitive = False
        destination_path = None
        prefix = None
        region = None
        source_url = None
        anonymous = None
        sha256 = False

        def run(self, kwargs):
            if kwargs is not None and len(kwargs) > 0:
                if not self.__parse_kwargs__(kwargs):
                    return
            self.__verify_arguments__()
            return self.download()

        def __parse_kwargs__(self, kwargs):
            if kwargs is None:
                return True
            if len(kwargs) == 0 and self.bucket_name is None and self.source_url is None:
                return self.help()
            for key, val in list(kwargs.items()):
                if key in HELP_KEYS:
                    return self.help()
                elif key in BUCKET_KEYS:
                    self.bucket_name = val
                elif key in CASE_INSENSITIVE_KEYS:
                    self.case_insensitive = True
                elif key in DESTINATION_KEYS:
                    self.destination_path = val
                elif key in PATH_KEYS:
                    self.prefix = val
                elif key in REGION_KEYS:
                    self.region = val
                elif key in SOURCE_KEYS:
                    self.source_url = val
                else:
                    print("Invalid option: " + str(val))
                    return False
                return True

        def __verify_arguments__(self):
            if self.bucket_name is None and self.source_url is None:
                raise DownloadError("You need to specify a bucket name or a source url.")
            if self.prefix is None:
                self.prefix = "/"
            # TODO: Add region mapping
            if self.region is None:
                self.region = 'us-east-1'
            if self.destination_path is None:
                self.destination_path = "./"

        def download(self):
            # Determine if we're parsing a url
            if self.bucket_name is None:
                self.bucket_name, self.prefix = parse_s3_url(self.source_url)
            # Connect to S3
            s3 = None
            try:
                s3 = get_s3_connection(anonymous=False)
                # Verify bucket
                verify_bucket(self.bucket_name, s3)
                self.anonymous = False
            except Exception:
                s3 = get_s3_connection(anonymous=True)
                self.anonymous = True
                # Verify bucket
                verify_bucket(self.bucket_name, s3)

            # Stupid s3 can't provide a length to their collections...
            count = 0
            # Return value
            destination_files = list()
            # Loop through found files
            for obj, checksum in find_files(self.bucket_name, self.prefix, case_sensitive=not self.case_insensitive, connection=s3, anonymous=self.anonymous, sha256=self.sha256):
                if obj.key.endswith("/"):
                    continue
                destination = self.destination_path
                # Case: Provided path exists
                if os.path.exists(destination):
                    # Case Provided path is a directory
                    if os.path.isdir(destination):
                        # Append file name to directory
                        destination = os.path.join(destination, os.path.split(obj.key)[1])
                    # Case Provided path is a file
                    else:
                        # TODO: do something
                        print ("Unconfirmed case")
                # Case: Provided path does not exist
                else:
                    # Case: Provided path ends with a path seperator
                    if destination.endswith(os.sep):
                        try:
                            # Make directories, and append file name
                            os.makedirs(destination)
                            destination = os.path.join(destination, os.path.split(obj.key)[1])
                        except OSError:
                            raise DownloadError("Unable to create directories for file: " + destination)
                    # Case: Provided path looks like it's a file
                    else:
                        # Check if parent directories exist, and if they don't, attempt to create them
                        head, tail = os.path.split(destination)
                        if not os.path.exists(head):
                            os.makedirs(head)
                # Perform download
                s3.meta.client.download_file(self.bucket_name, obj.key, destination)

                # Append downloaded file names
                destination_files.append((os.path.abspath(destination), checksum))

                # Increment counter
                count += 1

            if count == 0:
                raise DownloadError("No files found matching " + self.prefix)

            return destination_files


if __name__ == "__main__":
    # if __package__ is None:
    #     import sys
    #     from os import path
    #
    #     sys.path.append(path.dirname(path.dirname(path.abspath(__file__))))
    #     from maestro.core import module

    import time
    current_key = None
    keyvals = dict()

    for arg in sys.argv[1:]:
        if current_key is None:
            if arg.startswith('-'):
                current_key = arg.lstrip('-')
            continue
        if not arg.startswith("-"):
            keyvals[current_key] = arg
            current_key = None
        else:
            keyvals[current_key] = None
            current_key = arg.lstrip('-')
    if current_key is not None and current_key not in list(keyvals.keys()):
        keyvals[current_key] = ""

    s3dl = AsyncS3Downloader(None)
    s3dl.start(keyvals)
    while s3dl.status != module.DONE:
        if s3dl.exception is not None:
            raise s3dl.exception
        time.sleep(1)
        print ("Waiting...")
    if s3dl.exception is not None:
        raise s3dl.exception



"""
ioc.py

Contains simple container class for Modules and the associated Exceptions.

"""


class SingleObjectContainer(object):
    """
    Simple IoC Container.

    You can init with a type, and the container will only accept those. Otherwise it will take the object type of the first registered object.

    Don't change the object_type manually, and don't reuse it.

    """

    __objects = None
    object_type = None

    def __init__(self, obj_type=None):
        if self.object_type is not None and not isinstance(Type, obj_type):
            raise TypeError("Expected type Type got " + str(type(obj_type)) + ".")
        self.object_type = obj_type
        self.__objects = dict()

    def register(self, id, obj):
        if obj is None:
            raise TypeError("You cannot register a None type.")

        if id is None:
            raise TypeError("You cannot use a None type as an id.")

        if not isinstance(id, str):
            raise TypeError("You must pass a valid string for the id!")

        if self.object_type is None:
            self.object_type = type(obj)

        if not isinstance(obj, self.object_type):
            raise TypeError("The object passed to the Container was not of type " + str(self.object_type) + ".")

        if id in list(self.__objects.keys()):
            raise ValueError("ID " + str(id) + " is already registered with this container.")

        self.__objects[id] = obj

    def deregister(self, id):
        """
        Deregisters object with id. Will raise ValueError if the id does not exist.
        """

        if id not in list(self.__objects.keys()):
            raise ValueError("ID " + str(id) + " is not registered with this container.")

        del self.__objects[id]

    def getinstance(self, id):
        """
        Returns a new instance of an object with the provided id. The container does not track this object.
        """
        if id not in list(self.__objects.keys()):
            raise ValueError("ID " + str(id) + " is not registered with this container.")
        item_type = type(self[id])

        return item_type()

    def unregister(self, id):
        return self.deregister(id)

    def __getitem__(self, id):
        if id not in list(self.__objects.keys()):
            raise KeyError("A " + str(self.object_type) + " object with id " + str(id) + " is not registered with this container.")

        return self.__objects[id]

    def get(self, id):
        return self.__getitem__(id)

    def list(self):
        return list(self.__objects.items())



"""
parameter_store.py

AWS Parameter Store help methods - currently extending boto3

Note:
    Credentials are required to communicate with AWS.
    aws cli profile can be passed in using --profile, or
    the following ENVIRONMENT VARIABLES can be set before
    running this script:
        AWS_ACCESS_KEY_ID
        AWS_SECRET_ACCESS_KEY
"""

import logging
import argparse
import sys
import os
import boto3
import botocore


def _log_and_print_to_console(msg, log_level='info'):
    """
    Print a message to the console and log it to a file
    :param msg: the message to print and log
    :param log_level: the logging level for the mesage
    """
    log_func = {'info': logging.info, 'warn': logging.warn, 'error': logging.error}
    print(msg)
    log_func[log_level.lower()](msg)


def parse_file_into_string(file_path):
    with open(file_path) as f:
        return ''.join(f.readlines()).rstrip()


def get_param_from_region(requested_param, region, decrypt=False, profile=None):
    if not region:
        _log_and_print_to_console("ERROR: You must supply a region", 'error')
        sys.exit(1)

    return_value = None

    session = boto3.session.Session(profile_name=profile, region_name=region)
    ssm = session.client('ssm')

    result = ssm.get_parameters(Names=[requested_param], WithDecryption=decrypt)

    if result:
        if 'ResponseMetadata' in result:
            if 'HTTPStatusCode' in result['ResponseMetadata']:
                if result['ResponseMetadata']['HTTPStatusCode'] == 200:
                    if 'Parameters' in result:
                        for param in result['Parameters']:
                            if param['Name'] == requested_param:
                                return_value = param['Value']
                                break
    return return_value


def get_parameter(requested_param, region_list, decrypt=False, profile=None):
    result = {}

    for region in region_list:
        logging.debug("Checking region: " + region)
        result[region] = get_param_from_region(requested_param, region, decrypt, profile)
    return result


def set_param_in_region(region, parameter, value, value_is_file=False, description=None, encrypt=False, key=None, profile=None):
    if not region:
        _log_and_print_to_console("ERROR: You must supply a region", 'error')
        sys.exit(1)

    if not value:
        _log_and_print_to_console("ERROR: You must supply a value for the parameter", 'error')
        sys.exit(1)

    return_value = None

    session = boto3.session.Session(profile_name=profile, region_name=region)
    ssm = session.client('ssm')

    if encrypt:
        type = "SecureString"
    else:
        type = 'String'

    if value_is_file and not os.path.exists(value):
        _log_and_print_to_console("ERROR: File Value provided, but file does not exist", 'error')
        sys.exit(1)

    if value_is_file:
        value = parse_file_into_string(value)

    if description:
        if key:
            result = ssm.put_parameter(Name=parameter,
                                       Description=description,
                                       Value=value,
                                       Type=type,
                                       KeyId=key,
                                       Overwrite=True)
        else:
            result = ssm.put_parameter(Name=parameter,
                                       Description=description,
                                       Value=value,
                                       Type=type,
                                       Overwrite=True)
    else:
        if key:
            result = ssm.put_parameter(Name=parameter,
                                       Value=value,
                                       Type=type,
                                       KeyId=key,
                                       Overwrite=True)
        else:
            result = ssm.put_parameter(Name=parameter,
                                       Value=value,
                                       Type=type,
                                       Overwrite=True)

    if result:
        if 'ResponseMetadata' in result:
            if 'HTTPStatusCode' in result['ResponseMetadata']:
                return_value = result['ResponseMetadata']['HTTPStatusCode']
    return return_value


def set_paramameter(region_list, param_name, value, value_is_file=False, description=None, encrypt=False, key=None, profile=None):
    result = {}

    for region in region_list:
        logging.debug("Checking region: " + region)
        result[region] = set_param_in_region(region, param_name, value, value_is_file, description, encrypt, key, profile)
    return result


def delete_param_in_region(region, param_name, profile=None):
    if not region:
        _log_and_print_to_console("ERROR: You must supply a region", 'error')
        sys.exit(1)

    if not param_name:
        _log_and_print_to_console("ERROR: You must supply a parameter to delete", 'error')
        sys.exit(1)

    return_value = None

    session = boto3.session.Session(profile_name=profile, region_name=region)
    ssm = session.client('ssm')

    try:
        result = ssm.delete_parameter(Name=param_name)
    except botocore.exceptions.ClientError as e:
        if e.response['Error']['Code'] == 'ParameterNotFound':
            return 404
        else:
            _log_and_print_to_console('Unexpected error: %s' % e)
            return 404

    if result:
        if 'ResponseMetadata' in result:
            if 'HTTPStatusCode' in result['ResponseMetadata']:
                return_value = result['ResponseMetadata']['HTTPStatusCode']
    return return_value


def delete_paramameter(region_list, param_name, profile=None):
    result = {}

    for region in region_list:
        logging.debug("Checking region: " + region)
        result[region] = delete_param_in_region(region, param_name, profile)
    return result


if __name__ == "__main__":
    parser = argparse.ArgumentParser(description='parameter_store.py')

    me_cmd_group = parser.add_mutually_exclusive_group(required=True)
    me_cmd_group.add_argument("--get", help="Perform a Get", action="store_true")
    me_cmd_group.add_argument("--set", help="Perform a Set", action="store_true")
    me_cmd_group.add_argument("--delete", help="Perform a Delete", action="store_true")

    parser.add_argument("--param", help="Parameter name", dest='param', required=True)

    me_encrypt_group = parser.add_mutually_exclusive_group()
    me_encrypt_group.add_argument("--decrypt", help="Specify if parameter should be decrypted (for a GET)", action='store_true', default=False)
    me_encrypt_group.add_argument("--encrypt", help="Specify if parameter should be encrypted (for a SET)", action='store_true', default=False)

    me_value_group = parser.add_mutually_exclusive_group()
    me_value_group.add_argument("--value", help="Value of the parameter to set", dest='value')
    me_value_group.add_argument("--file-value", help="Use the contents of the given file for the value (multi-line not supported)", dest='file_value')

    parser.add_argument("--description", help="Description of the parameter to set", dest='description')
    parser.add_argument("--key", help="Parameter KeyID to use during set", dest='key')
    parser.add_argument("--regions", help="AWS Region(s) involved (space separated)", dest='region', nargs='+', required=True)
    parser.add_argument("--profile",
                        help="The name of an aws cli profile to use.", dest='profile', required=False)
    parser.add_argument("--verbose", help="Turn on DEBUG logging", action='store_true', required=False)
    parser.add_argument("--dryrun", help="Do a dryrun - no changes will be performed", dest='dryrun',
                        action='store_true', default=False,
                        required=False)

    args = parser.parse_args()

    log_level = logging.INFO

    if args.verbose:
        print("Verbose logging selected")
        log_level = logging.DEBUG

    logging.basicConfig(filename='parameter_store.log', format='%(asctime)s - %(levelname)7s : %(message)s',
                        level=log_level)

    logging.info("INIT")

    if args.get:
        result = get_parameter(args.param, args.region, args.decrypt, args.profile)
        for region in result:
            print(region + ': ' + (result[region] if result[region] else "Not Present"))

    if args.set:
        if not args.value and not args.file_value:
            _log_and_print_to_console("Must supply VALUE", "error")
            sys.exit(1)
        value_is_file = True if args.file_value else False
        param_value = args.file_value if args.file_value else args.value
        result = set_paramameter(args.region, args.param, param_value, value_is_file, args.description, args.encrypt, args.key, args.profile)
        for region in result:
            print(region + ': ' + ("Success" if result[region] == 200 else "Failed"))

    if args.delete:
        result = delete_paramameter(args.region, args.param, args.profile)
        for region in result:
            print(region + ': ' + ("Success" if result[region] == 200 else ("Not Found" if result[region] == 404 else "Failed")))



"""
Maestro File Tools
"""

import hashlib
import base64

def read_blocks(file, blocksize=1024):
    while True:
        block = file.read(blocksize)
        if block:
            yield block
        else:
            return

def sha256_simple(filename):
    m = hashlib.sha256()
    with open(filename, "rb") as f:
        # Read and update hash string value in blocks of 4K
        for byte_block in iter(lambda: f.read(8 * 1024 * 1024), b''):
            m.update(byte_block)
    return base64.b64encode(m.digest()).decode('utf-8')

def sha256_multipart(file_path, chunk_size=16 * 1024 * 1024):
    # chuck size of 16 mb works for our s3 downloads, multiplier may change if smaller chuck sizes
    sha256_hashes = []
    with open(file_path, 'rb') as fp:
        while True:
            data = fp.read(chunk_size)
            if not data:
                break
            sha256_hashes.append(hashlib.sha256(data))

    if len(sha256_hashes) == 1:
        return '"{}"'.format(sha256_hashes[0].hexdigest())

    digests = b''.join(sha.digest() for sha in sha256_hashes)
    digests_sha256 = hashlib.sha256(digests)
    return '{}-{}'.format(base64.b64encode(digests_sha256.digest()).decode('utf-8'), len(sha256_hashes))

def sha256_checksum(filename, checksum):
    checksum_stripped = checksum[1:-1]  # strip quotes
    if '-' in checksum_stripped:
        return sha256_multipart(filename)
    else:
        return sha256_simple(filename)

def md5_checksum(afile, blocksize=65536):
    with open(afile, "rb") as f:
        hasher = hashlib.md5()
        buf = f.read(blocksize)
        while len(buf) > 0:
            hasher.update(buf)
            buf = f.read(blocksize)
        return hasher.hexdigest()



import os
import platform
import re
import shutil


def get_tree_size(path='.'):
    """
    get_tree_size will return the total size of a directory tree
    """
    if not os.path.exists(path):
        raise OSError("Path " + str(path) + " does not exist!")

    total_size = 0
    for dirpath, dirnames, filenames in os.walk(str(path)):
        for f in filenames:
            fp = os.path.join(dirpath, f)
            total_size += os.path.getsize(fp)
    return total_size


def get_case_insensitive_path(path='.'):
    """
    get_case_insensitive_path will check for the existance of a path in a case sensitive file system, regardless of the case of the inputted path. Returns the absolute path if found (with correct casing) or None.
    """
    if os.path.exists(path):
        return path
    elif platform.system() == "Windows":
        return None

    path_elements = full_split(path)
    path_root = None

    drive, path = os.path.splitdrive(path)
    if not drive:
        if not path.startswith("/"):
            path_root = os.path.abspath(os.path.normpath("./"))
        else:
            path_root = os.path.abspath(os.path.normpath("/"))
    else:
        path_root = os.path.abspath(os.path.normpath(drive))

    if not os.path.exists(path_root):
        raise OSError("Unable to locate path root: " + str(path_root))

    # Build the full path, also used for error messages
    full_path = path_root
    for element in path_elements:
        if not element or element == "/" or element == ".":
            continue
        found = False
        for directory in os.listdir(full_path):
            if element.lower() == directory.lower():
                full_path = os.path.join(full_path, directory)
                found = True
                break
        if found is False:
            return None
    return full_path


# Credit: Gian Marco Gherardi
#         http://stackoverflow.com/questions/6260149/os-symlink-support-in-windows
def symlink(source, link_name):
    import os
    os_symlink = getattr(os, "symlink", None)
    if callable(os_symlink):
        os_symlink(source, link_name)
    else:
        import ctypes
        csl = ctypes.windll.kernel32.CreateSymbolicLinkW
        csl.argtypes = (ctypes.c_wchar_p, ctypes.c_wchar_p, ctypes.c_uint32)
        csl.restype = ctypes.c_ubyte
        flags = 1 if os.path.isdir(source) else 0
        if csl(link_name, source, flags) == 0:
            raise ctypes.WinError()


def purge(pattern, path, match_directories=False):
    for root, dirs, files in os.walk(path):
        if match_directories is True:
            for dir in [x for x in dirs if re.match(pattern, x)]:
                shutil.rmtree(os.path.join(root, dir))
        for file in [x for x in files if re.match(pattern, x)]:
            os.remove(os.path.join(root, file))


# Credit: John Machin
#         http://stackoverflow.com/questions/4579908/cross-platform-splitting-of-path-in-python
def full_split(path, debug=False):
    """
    full_split will split Windows and UNIX paths into seperate elements
    """
    parts = []
    while True:
        newpath, tail = os.path.split(path)
        if debug:
            print (repr(path), (newpath, tail))
        if newpath == path:
            assert not tail
            if path:
                parts.append(path)
            break
        parts.append(tail)
        path = newpath
    parts.reverse()
    return parts



import re


def replaceall(replace_dict, string):
    """
    replaceall will take the keys in replace_dict and replace string with their corresponding values. Keys can be regular expressions.
    """
    replace_dict = dict((re.escape(k), v) for k, v in list(replace_dict.items()))
    pattern = re.compile("|".join(list(replace_dict.keys())))
    return pattern.sub(lambda m: replace_dict[re.escape(m.group(0))], string)



"""
module.py

Modules are one-off runnable tasks. All modules should extend the Module() class.

Modules are meant to help build modular tools.
"""
import types
import copyreg
import sys
import traceback
import multiprocessing
from multiprocessing.pool import Pool
from logging import INFO

# Module states
NOT_STARTED = 0
RUNNING = 1
DONE = 2
PROCESSED = sys.maxsize


# Exceptions
class AsyncException(Exception):
    """
    Async exceptions are the exceptions that are returned from an AsyncModule. It provides the
    original traceback as a string under "traceback", and also provides the original string
    message under "message"
    """
    traceback = None


class NoDaemonProcess(multiprocessing.Process):
    @property
    def daemon(self):
        return False

    @daemon.setter
    def daemon(self, value):
        pass


class NoDaemonContext(type(multiprocessing.get_context())):
    Process = NoDaemonProcess


class NonDaemonizedPool(Pool):
    def __init__(self, *args, **kwargs):
        kwargs['context'] = NoDaemonContext()
        super(NonDaemonizedPool, self).__init__(*args, **kwargs)

# TODO: change modules to accept kwargs and args


class Module(object):
    """
    Base module class. All modules may take an ObjectContainer on initialization, and must override the run() method
    """
    id = ''
    __ioc__ = None

    def __init__(self, ioc=None):
        self.__ioc__ = ioc

    def getObject(self, name):
        return self.__ioc__.get(name)

    def getObjectInstance(self, name):
        return self.__ioc__.getinstance(name)

    def run(self, kwargs={}):
        pass

    def start(self, kwargs={}):
        self.run(kwargs=kwargs)

    def help(self):
        try:
            print(self.HELPTEXT)
        except NameError:
            print("No help defined for module " + str(type(self)))
        return True


def __pickle_method__(m):
    """
    Pickling override for using pickle with instance methods.
    """
    if m.__self__ is None:
        return getattr, (m.__self__.__class__, m.__func__.__name__)
    else:
        return getattr, (m.__self__, m.__func__.__name__)


copyreg.pickle(types.MethodType, __pickle_method__)


class AsyncModule(Module):
    """
    Builds on the Module class to add asynchronous functionality for modules that can support it.

    Intra/inter synchronization must be handled by each module. Modules are seperate processes.
    """
    status = None
    result = None
    exception = None
    __logger__ = None

    def __init__(self, ioc=None):
        """
        Sets the current status and calls the superclass' init
        """
        super(AsyncModule, self).__init__(ioc=ioc)
        self.status = NOT_STARTED

    def start(self, kwargs={}):
        """
        The main method to start a module. In Async, it will return immediately with the result from apply_async.
        """
        pool = NonDaemonizedPool(processes=1)
        self.status = RUNNING
        result = pool.apply_async(self.__setup__, [kwargs], callback=self.__finish_internal__)
        return result

    def __setup__(self, kwargs):
        """
        Internal setup method, this should not be overridden unless you know what you're doing. Calls the run method.
        """
        # If we're an AsyncModule, we need to catch Exceptions so they can be converted and passed to the parent process
        try:
            return self.run(kwargs=kwargs)
        except Exception as e:
            try:
                exc = AsyncException(e)
                exc.traceback = "".join(traceback.format_exception(*sys.exc_info()))
                return exc
            except Exception as e:
                print("FATAL ERROR -- " + str(e))

    def __finish_internal__(self, callback_args):
        """
        Internal finish method. This should not be overridden unless you know what you're doing. Calls the finish method.
        """
        self.status = DONE
        self.finish(callback_args)

    def finish(self, results):
        """
        Finish method, overridable but otherwise sticks the results into a variable, or sets the exception.
        """
        if isinstance(results, Exception):
            self.exception = results
        else:
            self.result = results

    def log(self, message, level=INFO):
        """
        Log method, currently set to only print to the console. Thread-safe.
        """
        if self.__logger__ is None:
            self.__logger__ = multiprocessing.log_to_stderr()
            self.__logger__.setLevel(INFO)
        self.__logger__.info(message)



import os
import sys
import subprocess


def check_pid(process_id):
    if sys.platform == "win32":
        pids = []
        tasklist = subprocess.Popen(["tasklist"], stdout=subprocess.PIPE)
        try:
            for task in tasklist.communicate()[0].split("\r\n"):
                try:
                    pids.append(int(task[26:34]))
                except Exception:
                    pass
        except Exception as e:
            raise OSError("Unable to list process ids: " + str(e))
        if int(process_id) in pids:
            return True
        else:
            return False
    else:
        try:
            os.kill(process_id, 0)
        except OSError:
            return False
        else:
            return True



"""
r53.py

Route 53 help methods - currently extending boto3

Note:
    Credentials are required to communicate with AWS.
    aws cli profile can be passed in using --profile, or
    the following ENVIRONMENT VARIABLES can be set before
    running this script:
        AWS_ACCESS_KEY_ID
        AWS_SECRET_ACCESS_KEY
"""

import boto3
import botocore
import argparse
import logging

logging.getLogger("botocore").setLevel(logging.CRITICAL)


def get_hosted_zone_list(profile=None):
    session = boto3.session.Session(profile_name=profile)
    r53_client = session.client('route53')
    result = None
    try:
        query = r53_client.list_hosted_zones()
        result = query['HostedZones']
    except botocore.exceptions.ClientError as e:
        logging.error('Unexpected error: %s' % e)
    return result


def get_all_records_in_zone(zone_id, profile=None):
    session = boto3.session.Session(profile_name=profile)
    r53_client = session.client('route53')
    resource_record_sets = []
    try:
        current_set = r53_client.list_resource_record_sets(HostedZoneId=zone_id)
        resource_record_sets.extend(current_set['ResourceRecordSets'])
        isTruncated = current_set['IsTruncated']
        while isTruncated:
            start_name = current_set['NextRecordName']
            current_set = r53_client.list_resource_record_sets(HostedZoneId=zone_id, StartRecordName=start_name)
            resource_record_sets.extend(current_set['ResourceRecordSets'])
            isTruncated = current_set['IsTruncated']
    except botocore.exceptions.ClientError as e:
        logging.error('Unexpected error: %s' % e)
    return resource_record_sets


def get_record_in_zone(zone_id, record_name, profile=None):
    record_list = get_all_records_in_zone(zone_id, profile)
    wildcard = False
    if '*' in record_name:
        record_name = record_name.strip('*')
        wildcard = True
    else:
        if not record_name.endswith('.'):
            record_name += '.'
    result = []
    for record in record_list:
        if wildcard:
            if record_name in record['Name'].split('.')[0]:
                result.append(record)
        else:
            if record_name == record['Name']:
                result.append(record)
    return result


def get_hosted_zone_by_id(zone_id, profile=None):
    session = boto3.session.Session(profile_name=profile)
    r53_client = session.client('route53')
    result = None
    try:
        query = r53_client.get_hosted_zone(Id=zone_id)
        result = query['HostedZone']
    except botocore.exceptions.ClientError as e:
        if e.response['Error']['Code'] == 'NoSuchHostedZone':
            logging.error("Zone does not exist")
        else:
            logging.error('Unexpected error: %s' % e)
    return result


def get_hosted_zone_by_name(zone_name, profile=None):
    session = boto3.session.Session(profile_name=profile)
    r53_client = session.client('route53')
    result = None
    if not zone_name.endswith('.'):
        zone_name += '.'
    zone_list = get_hosted_zone_list(profile=profile)
    if zone_list:
        for zone in zone_list:
            if zone_name == zone['Name']:
                try:
                    query = r53_client.get_hosted_zone(Id=zone['Id'])
                    result = query['HostedZone']
                    break
                except botocore.exceptions.ClientError as e:
                    if e.response['Error']['Code'] == 'NoSuchHostedZone':
                        logging.error("Zone does not exist")
                    else:
                        logging.error('Unexpected error: %s' % e)
    return result


def zone_exists(zone_name, profile=None):
    zone_info = get_hosted_zone_by_name(zone_name, profile)
    if zone_info:
        return True
    else:
        return False


def record_as_fqdn(record_name, zone):
    record = record_name.split('.')
    if len(record) < 2:
        record_name = record_name + '.' + zone
    return record_name


def record_exists(record_name, zone_id=None, profile=None):
    alias = record_name
    if not zone_id:
        # Zone Not provided - figure it out
        record = record_name.split('.')
        if len(record) < 2:
            logging.error("Must supply FQDN if no zone provided")
            return False
        else:
            alias = record[0]
            zone = '.'.join(record[1:])
            zone_info = get_hosted_zone_by_name(zone, profile)
            if not zone_info:
                return False
            else:
                zone_id = zone_info['Id']
    else:
        if len(record_name.split('.')) > 1:
            # Record name to find is the first element
            alias = record_name.split('.')[0]
    records = get_all_records_in_zone(zone_id, profile)
    for record in records:
        if record['Name'].split('.')[0] == alias:
            # Found it
            return True
    return False


def _print_hosted_zone(zone_info):
    for key in zone_info:
        print("%23s: %s" % (key, zone_info[key]))
    print('')


def _print_resource_record(record):
    for key in record:
        print("%15s: %s" % (key, record[key]))
    print('')


if __name__ == "__main__":

    parser = argparse.ArgumentParser(
        description='Script to view/modify Route 53 entries')

    me_cmd_group = parser.add_mutually_exclusive_group(required=True)
    me_cmd_group.add_argument("--get", help="get Route 53 record(s)/zone(s) - wildcards supported", action='store_true')
    me_cmd_group.add_argument("--exists", help="determine if Route 53 record(s)/zone(s) exist", action='store_true')
    # TODO: --set, --delete, --update
    # me_cmd_group.add_argument("--set", help="set Route 53 record", action='store_true')
    # me_cmd_group.add_argument("--delete", help="delete Route 53 record", action='store_true')
    # me_cmd_group.add_argument("--update", help="update Route 53 record", action='store_true')

    parser.add_argument("--zones", help="The name of the hosted zone(s) (all is valid)",
                        dest='zones', nargs='+', metavar='NAME')
    parser.add_argument("--records", help="The name of the resource record(s) (all is valid)",
                        dest='records', nargs='+', metavar='NAME')

    parser.add_argument("--profile",
                        help="The name of an aws cli profile to use.", dest='profile', required=False)
    parser.add_argument("--verbose", help="Turn on DEBUG logging", action='store_true', required=False)
    args = parser.parse_args()

    log_level = logging.INFO

    if args.verbose:
        log_level = logging.DEBUG

    logger = logging.getLogger()
    logger.setLevel(logging.DEBUG)
    # create file handler which logs even debug messages
    fh = logging.FileHandler('r53.log')
    fh.setLevel(logging.DEBUG)
    # create console handler using level set in log_level
    ch = logging.StreamHandler()
    ch.setLevel(log_level)
    console_formatter = logging.Formatter('%(levelname)8s: %(message)s')
    ch.setFormatter(console_formatter)
    file_formatter = logging.Formatter('%(asctime)s - %(levelname)8s: %(message)s')
    fh.setFormatter(file_formatter)
    # Add the handlers to the logger
    logger.addHandler(fh)
    logger.addHandler(ch)

    logging.debug("INIT")

    if not args.zones and not args.records:
        logging.error("Must supply either zones, records or both")

    if args.get:
        if not args.zones or 'all' in args.zones:
            # All zones requested - get the zone list
            zone_list = get_hosted_zone_list(profile=args.profile)
            for zone in zone_list:
                if args.records:
                    if 'all' in args.records:
                        # get all records for this zone
                        print("Getting all records for %s" % zone['Name'])
                        record_list = get_all_records_in_zone(zone['Id'], profile=args.profile)
                        for record in record_list:
                            _print_resource_record(record)
                    else:
                        # Get specific records in all zones
                        for record in args.records:
                            if '*' not in record:
                                record = record_as_fqdn(record, zone['Name'])
                            record_info = get_record_in_zone(zone['Id'], record, profile=args.profile)
                            for r in record_info:
                                _print_resource_record(r)
                else:
                    # No records requested - just print the zone
                    print(zone['Name'])
        else:
            # Specific zone(s) requested
            for zone in args.zones:
                zone_info = get_hosted_zone_by_name(zone, profile=args.profile)
                if zone_info:
                    if args.records:
                        if 'all' in args.records:
                            print("Getting all records for %s" % zone_info['Name'])
                            record_list = get_all_records_in_zone(zone_info['Id'], profile=args.profile)
                            for record in record_list:
                                _print_resource_record(record)
                        else:
                            # Get specific records in specifc zones
                            for record in args.records:
                                if '*' not in record:
                                    record = record_as_fqdn(record, zone_info['Name'])
                                record_info = get_record_in_zone(zone_info['Id'], record, profile=args.profile)
                                for r in record_info:
                                    _print_resource_record(r)
                    else:
                        # No records requested - just get zone info
                        _print_hosted_zone(zone_info)
                else:
                    logging.error("%s does not exist" % zone)
    elif args.exists:
        if (args.zones and 'all' in args.zones) or (args.records and 'all' in args.records):
            logging.error("all not supported with exists")

        if args.zones and args.records:
            # Looking for specific records in specific zones
            results = {}
            for zone in args.zones:
                if zone_exists(zone, profile=args.profile):
                    zone_info = get_hosted_zone_by_name(zone, profile=args.profile)
                    results[zone] = {}
                    for record in args.records:
                        results[zone][record] = record_exists(record, zone_info['Id'], profile=args.profile)
                else:
                    results[zone] = None
            for result in results:
                print("Zone: %s" % result)
                if results[result]:
                    for record in results[result]:
                        print ('   ' + record + ': ' + str(results[result][record]))
                else:
                    logging.error("   Zone doesn't exist")

        elif args.zones and not args.records:
            # Looking for specific zones
            results = {}
            for zone in args.zones:
                results[zone] = zone_exists(zone, profile=args.profile)
            for result in results:
                print(result + ': ' + str(results[result]))

        elif not args.zones and args.records:
            # Looking for specific records (without supplying zone info)
            results = {}
            for record in args.records:
                results[record] = record_exists(record, None, profile=args.profile)
            for result in results:
                print(result + ': ' + str(results[result]))



from setuptools import setup, find_packages
from distutils.core import setup

setup(name='MaestroOps',
      version='0.9.1',
      description='Python Automation Framework for Development Operations Teams',
      author='Signiant DevOps',
      author_email='devops@signiant.com',
      url='https://www.signiant.com',
      packages=find_packages(),
      license='MIT',
      install_requires=['boto3>=1.33.1,<1.34.0']
     )



"""
jobs.py

Contains various definitions of Jenkins jobs. All job types should be based off
of a JenkinsJobEntry.

The goal would be to have a full config parser, possibly using a factory if the performance of parsing a full config file for a few attributes is terrible.
"""

# TODO:
# Make verbose/debug useful (they used to be)
# Expand on config parsing

import os
from xml.dom import minidom
from glob import glob

JENKINS_DEFAULT_JOB_CONFIG_FILE = "config.xml"


class InvalidEntryError(Exception):
    pass


class JenkinsJobEntry(object):
    """
    Contains basic information about a Jenkins job. All information parsed by this class comes from the folder structure, but that may change in the future.
    """
    # The location of the Jenkins build directory (/var/lib/jenkins/jobs/.../builds)
    build_path = None

    # The jenkins config file location
    config_file_path = None

    # The root folder of Jenkins (currently unpopulated)
    jenkins_root = None

    # Jenkins Job Name
    name = None

    def __init__(self, job_directory, verbose=False, debug=False):
        """
        Init verifies that the job_directory does indeed seem like a Jenkins job folder. If the config file is not found it will raise an exception.
        """

        if job_directory is None:
            raise TypeError("Unable to process a None type.")

        self.config_file_path = os.path.join(job_directory, JENKINS_DEFAULT_JOB_CONFIG_FILE)
        if not os.path.exists(self.config_file_path):
            raise InvalidEntryError("The provided job does not contain a " + JENKINS_DEFAULT_JOB_CONFIG_FILE + " file under " + str(job_directory))

        self.build_path = os.path.join(job_directory, "builds")

        self.name = os.path.basename(job_directory)

    def get_build_number_list(self, verbose=False, debug=False):

        # Get all numeric folder names in the builds folder, assigning it to self is a cheap way to cache it
        self.builds_in_jenkins = [os.path.basename(f) for f in glob(self.build_path + "/[0-9]*") if os.path.isdir(f)]

        if verbose is True:
            for b in self.builds_in_jenkins:
                print ("Found build " + b + " in Jenkins")
        return self.builds_in_jenkins


def parse_build_into_environment_variable_job_entry(build_root):
    env_file = os.path.abspath(os.path.join(build_root, "builds", "lastSuccessfulBuild", "injectedEnvVars.txt"))
    entry = EnvironmentVariableJobEntry(build_root)
    entry.environment_variables = dict()
    try:
        with open(env_file, "r") as f:
            for line in f.readlines():
                try:
                    elements = line.strip().split("=")
                    entry.environment_variables[elements[0]] = elements[1]
                except IndexError:
                    continue
    except IOError:
        pass
        # print "Unable to open the environment variabiles file: " + str(env_file)
    return entry


class EnvironmentVariableJobEntry(JenkinsJobEntry):
    """
    Represents the  metadata of a Jenkin's job which uses the envinject plugin from Signiant's main Jenkins server

    Ideally we would want to seperate the xml parsing of the jenkins out and create a "config parser" type of thing
    and create a generic "JenkinsJobEntry", but it's so far from actually parsing the entire file it's not worth it
    until we need to parse more of the file.
    """

    # Environment variables dictionary
    environment_variables = None

    # This should be moved up into the base class, and we should probably make a factory
    disabled = None

    def __init__(self, job_directory, verbose=False, debug=False):

        # Call the superclass' init
        self.super = super(EnvironmentVariableJobEntry, self)
        self.super.__init__(job_directory, verbose, debug)

        try:
            self.__parse_environment_variables(self.config_file_path, verbose, debug)

        # Fatal Exceptions due to general XML Parsing
        except(TypeError, AttributeError):
            raise InvalidEntryError("Unable to parse the XML for this entry!")

        # Fatal Exception due to not having a plugin attribute
        except(KeyError):
            if debug is True:
                print "ERROR: It appears that the job using " + self.config_file_path + " does not have a section for environment variables."
            raise InvalidEntryError("Unable to find a plugin properties node")

    def __parse_environment_variables(self, config_file, verbose, debug):
        if debug is True:
            print ("Parsing " + config_file)
        # Parse initial xml document
        xml_document = minidom.parse(config_file)

        # Get Disabled Value (wtf XML syntax)
        disabled_exists = xml_document.getElementsByTagName("disabled")
        disabled_val = None
        if disabled_exists:
            disabled_val = xml_document.getElementsByTagName("disabled")[0].firstChild.nodeValue

        # Determine if it's disabled or not
        if disabled_val:
            if disabled_val == "True" or disabled_val == "true":
                self.disabled = True
            else:
                self.disabled = False
        else:
            self.disabled = False

        # Declare the dictionary so it's in scope
        environment_variables_dict = dict()

        # Find the envinject node to get the environment variables
        for node in xml_document.getElementsByTagName("propertiesContent"):
            node_parent_parent = node.parentNode.parentNode
            if "envinject" not in str(node_parent_parent.attributes["plugin"].value):
                continue

            # The actual text inside this node is considered a "TEXT_NODE", so we extract that with the following statement
            environment_variables_string = " ".join(t.nodeValue for t in node.childNodes if t.nodeType == t.TEXT_NODE)

            # Parse the properties into a dictionary
            for item in environment_variables_string.splitlines():
                try:
                    key, val = item.split("=")
                    environment_variables_dict[key] = val
                except ValueError as e:
                    if debug:
                        print ("WARNING: Hit empty entry, continuing.")
                        print (str(e))
                    pass

        if len(environment_variables_dict) == 0:
            if debug is True:
                print ("ERROR: It appears that the job using " + config_file + " does not have a section for environment variables.")
            raise InvalidEntryError("Unable to find the envinject properties node")

        self.environment_variables = environment_variables_dict



"""
ssm_documents.py

AWS SSM Documents help methods - currently extending boto3

Note:
    Credentials are required to communicate with AWS.
    aws cli profile can be passed in using --profile, or
    the following ENVIRONMENT VARIABLES can be set before
    running this script:
        AWS_ACCESS_KEY_ID
        AWS_SECRET_ACCESS_KEY
"""

import logging
import argparse
import sys
import os
import json
import boto3
import botocore


def _log_and_print_to_console(msg, log_level='info'):
    """
    Print a message to the console and log it to a file
    :param msg: the message to print and log
    :param log_level: the logging level for the mesage
    """
    log_func = {'info': logging.info, 'warn': logging.warn, 'error': logging.error}
    print(msg)
    log_func[log_level.lower()](msg)


def parse_file_into_json_string(file_path):
    with open(file_path) as f:
        return json.dumps(json.load(f))


def get_document_from_region(requested_document, region, profile=None):
    if not region:
        _log_and_print_to_console("ERROR: You must supply a region", 'error')
        sys.exit(1)

    return_value = None

    session = boto3.session.Session(profile_name=profile, region_name=region)
    ssm = session.client('ssm')

    try:
        result = ssm.get_document(Name=requested_document)
        if result:
            if 'Content' in result:
                return_value = result['Content']
    except botocore.exceptions.ClientError as e:
        if e.response['Error']['Code'] == 'InvalidDocument':
            return None
        else:
            _log_and_print_to_console('Unexpected error: %s' % e, 'error')

    return return_value


def get_document(region_list, requested_document, profile=None):
    result = {}

    for region in region_list:
        logging.debug("Checking region: " + region)
        result[region] = get_document_from_region(requested_document, region, profile)
    return result


def set_document_in_region(region, document_name, type, content, content_is_file=False, profile=None):
    if not region:
        _log_and_print_to_console("ERROR: You must supply a region", 'error')
        return False

    if not type:
        _log_and_print_to_console("ERROR: You must supply a type for the document - Command | Policy | Automation", 'error')
        return False

    if not content:
        _log_and_print_to_console("ERROR: You must supply content for the document", 'error')
        return False

    return_value = False

    session = boto3.session.Session(profile_name=profile, region_name=region)
    ssm = session.client('ssm')

    if content_is_file and not os.path.exists(content):
        _log_and_print_to_console("ERROR: File Value provided, but file does not exist", 'error')
        return False

    if content_is_file:
        content = parse_file_into_json_string(content)

    result = ssm.create_document(Name=document_name,
                                 Content=content,
                                 DocumentType=type)

    if result:
        if 'DocumentDescription' in result:
            # TODO: Should probably check the Status and react accordingly, but for now, as long as we have a
            #       DocumentDescription, assume all went well
            return_value = True
    return return_value


def set_document(region_list, document_name, type, content, content_is_file=False, profile=None):
    result = {}

    for region in region_list:
        logging.debug("Checking region: " + region)
        result[region] = set_document_in_region(region, document_name, type, content, content_is_file, profile)
    return result


def update_document_in_region(region, document_name, content, version=None, content_is_file=False, profile=None):
    if not region:
        _log_and_print_to_console("ERROR: You must supply a region", 'error')
        return False

    if not content:
        _log_and_print_to_console("ERROR: You must supply content for the document", 'error')
        return False

    return_value = False

    session = boto3.session.Session(profile_name=profile, region_name=region)
    ssm = session.client('ssm')

    if content_is_file and not os.path.exists(content):
        _log_and_print_to_console("ERROR: File Value provided, but file does not exist", 'error')
        return False

    if content_is_file:
        content = parse_file_into_json_string(content)

    if not version:
        version = '$LATEST'

    try:
        result = ssm.update_document(Name=document_name,
                                     Content=content,
                                     DocumentVersion=version)
    except botocore.exceptions.ClientError as e:
        if e.response['Error']['Code'] == 'DuplicateDocumentContent':
            return True
        else:
            _log_and_print_to_console('Unexpected error: %s' % e, 'error')

    if result:
        if 'DocumentDescription' in result:
            # TODO: Should probably check the Status and react accordingly, but for now, as long as we have a
            #       DocumentDescription, assume all went well
            return_value = True
    return return_value


def update_document(region_list, document_name, content, version=None, content_is_file=False, profile=None):
    result = {}

    for region in region_list:
        logging.debug("Checking region: " + region)
        result[region] = update_document_in_region(region, document_name, content, version, content_is_file, profile)
    return result


def delete_document_in_region(region, document_name, profile=None):
    if not region:
        _log_and_print_to_console("ERROR: You must supply a region", 'error')
        return False

    if not document_name:
        _log_and_print_to_console("ERROR: You must supply a parameter to delete", 'error')
        return False

    return_value = False

    session = boto3.session.Session(profile_name=profile, region_name=region)
    ssm = session.client('ssm')

    try:
        result = ssm.delete_document(Name=document_name)
        if 'ResponseMetadata' in result:
            if 'HTTPStatusCode' in result['ResponseMetadata']:
                if result['ResponseMetadata']['HTTPStatusCode'] == 200:
                    return True
                else:
                    return False
    except botocore.exceptions.ClientError as e:
        if e.response['Error']['Code'] == 'InvalidDocument':
            _log_and_print_to_console('Document does not exist', 'warn')
            return False
        else:
            _log_and_print_to_console('Unexpected error: %s' % e)
            return False

    if result:
        if 'ResponseMetadata' in result:
            if 'HTTPStatusCode' in result['ResponseMetadata']:
                return_value = result['ResponseMetadata']['HTTPStatusCode']
    return return_value


def delete_document(region_list, document_name, profile=None):
    result = {}

    for region in region_list:
        logging.debug("Checking region: " + region)
        result[region] = delete_document_in_region(region, document_name, profile)
    return result


if __name__ == "__main__":
    parser = argparse.ArgumentParser(description='parameter_store.py')

    me_cmd_group = parser.add_mutually_exclusive_group(required=True)
    me_cmd_group.add_argument("--get", help="Perform a Get", action="store_true")
    me_cmd_group.add_argument("--set", help="Perform a Set", action="store_true")
    me_cmd_group.add_argument("--update", help="Perform an Update", action="store_true")
    me_cmd_group.add_argument("--delete", help="Perform a Delete", action="store_true")

    parser.add_argument("--name", help="Document name", dest='name', required=True)
    parser.add_argument("--type", help="Document type", dest='type')
    parser.add_argument("--version", help="Document version, if unspecified, '$LATEST' is used", dest='version')

    me_value_group = parser.add_mutually_exclusive_group()
    me_value_group.add_argument("--content", help="Document content. NOTE: should be a JSON string", dest='content')
    me_value_group.add_argument("--file-content", help="Use the contents of the given file for the content. NOTE: file contents should be valid JSON", dest='file_content')

    parser.add_argument("--regions", help="AWS Region(s) involved (space separated)", dest='regions', nargs='+', required=True)
    parser.add_argument("--profile",
                        help="The name of an aws cli profile to use.", dest='profile', required=False)
    parser.add_argument("--verbose", help="Turn on DEBUG logging", action='store_true', required=False)
    parser.add_argument("--dryrun", help="Do a dryrun - no changes will be performed", dest='dryrun',
                        action='store_true', default=False,
                        required=False)

    args = parser.parse_args()

    log_level = logging.INFO

    if args.verbose:
        print("Verbose logging selected")
        log_level = logging.DEBUG

    logging.basicConfig(filename='ssm_documents.log', format='%(asctime)s - %(levelname)7s : %(message)s',
                        level=log_level)

    logging.info("INIT")

    if args.get:
        result = get_document(args.regions, args.name, args.profile)
        for region in result:
            print(region + ': ' + ('\n' + result[region] if result[region] else "Not Present"))

    if args.set:
        if not args.type:
            _log_and_print_to_console("Must supply TYPE", "error")
            sys.exit(1)
        if not args.content and not args.file_content:
            _log_and_print_to_console("Must supply CONTENT", "error")
            sys.exit(1)
        content_is_file = True if args.file_content else False
        content = args.file_content if args.file_content else args.content
        result = set_document(args.regions, args.name, args.type, content, content_is_file, args.profile)
        for region in result:
            print(region + ': ' + ("Success" if result[region] else "Failed"))

    if args.update:
        if not args.content and not args.file_content:
            _log_and_print_to_console("Must supply CONTENT", "error")
            sys.exit(1)
        content_is_file = True if args.file_content else False
        content = args.file_content if args.file_content else args.content
        result = update_document(args.regions, args.name, content, args.version, content_is_file, args.profile)
        for region in result:
            print(region + ': ' + ("Success" if result[region] else "Failed"))

    if args.delete:
        result = delete_document(args.regions, args.name, args.profile)
        for region in result:
            print(region + ': ' + ("Success" if result[region] else "Failed"))
